@@ -193,6 +193,30 @@ void to_utf8_test()
     }
 }
 
+static void print_utf16(unsigned long val)
+{
+    std::string result = QUtil::toUTF16(val);
+    std::cout << "0x" << QUtil::int_to_string_base(val, 16) << " ->";
+    for (std::string::iterator iter = result.begin();
+         iter != result.end(); ++iter)
+    {
+        std::cout << " " << QUtil::int_to_string_base(
+            static_cast<int>(static_cast<unsigned char>(*iter)), 16, 2);
+    }
+    std::cout << std::endl;
+}
+
+void to_utf16_test()
+{
+    print_utf16(0x41UL);
+    print_utf16(0xF7UL);
+    print_utf16(0x3c0UL);
+    print_utf16(0x16059UL);
+    print_utf16(0xdeadUL);
+    print_utf16(0x7fffffffUL);
+    print_utf16(0x80000000UL);
+}
+
 void print_whoami(char const* str)
 {
     PointerHolder<char> dup(true, QUtil::copy_string(str));
@@ -289,22 +313,25 @@ int main(int argc, char* argv[])
 {
     try
     {
+        std::cout << "---- string conversion" << std::endl;
 	string_conversion_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- os wrapper" << std::endl;
 	os_wrapper_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- fopen" << std::endl;
 	fopen_wrapper_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- getenv" << std::endl;
 	getenv_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- utf8" << std::endl;
 	to_utf8_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- utf16" << std::endl;
+	to_utf16_test();
+	std::cout << "---- whoami" << std::endl;
 	get_whoami_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- file" << std::endl;
 	same_file_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- lines from file" << std::endl;
 	read_lines_from_file_test();
-	std::cout << "----" << std::endl;
+	std::cout << "---- hex encode/decode" << std::endl;
 	hex_encode_decode_test();
     }
     catch (std::exception& e)
