@@ -3,6 +3,9 @@
 
 #include <qpdf/QPDF.hh>
 
+#include <qpdf/QPDFPageDocumentHelper.hh>
+#include <qpdf/QPDFPageObjectHelper.hh>
+#include <qpdf/QPDFAcroFormDocumentHelper.hh>
 #include <qpdf/QUtil.hh>
 #include <qpdf/QTC.hh>
 #include <qpdf/Pl_StdioFile.hh>
@@ -151,6 +154,13 @@ static QPDFObjectHandle createPageContents(QPDF& pdf, std::string const& text)
     return QPDFObjectHandle::newStream(&pdf, contents);
 }
 
+static void print_rect(std::ostream& out,
+                       QPDFObjectHandle::Rectangle const& r)
+{
+    out << "[" << r.llx << ", " << r.lly << ", "
+        << r.urx << ", " << r.ury << "]";
+}
+
 void runtest(int n, char const* filename1, char const* arg2)
 {
     // Most tests here are crafted to work on specific files.  Look at
@@ -455,12 +465,13 @@ void runtest(int n, char const* filename1, char const* arg2)
     }
     else if (n == 5)
     {
-	std::vector<QPDFObjectHandle> pages = pdf.getAllPages();
+        QPDFPageDocumentHelper dh(pdf);
+	std::vector<QPDFPageObjectHelper> pages = dh.getAllPages();
 	int pageno = 0;
-	for (std::vector<QPDFObjectHandle>::iterator iter = pages.begin();
+	for (std::vector<QPDFPageObjectHelper>::iterator iter = pages.begin();
 	     iter != pages.end(); ++iter)
 	{
-	    QPDFObjectHandle& page = *iter;
+	    QPDFPageObjectHelper& page(*iter);
 	    ++pageno;
 
 	    std::cout << "page " << pageno << ":" << std::endl;
@@ -633,11 +644,13 @@ void runtest(int n, char const* filename1, char const* arg2)
     }
     else if (n == 10)
     {
-	std::vector<QPDFObjectHandle> pages = pdf.getAllPages();
-	pages.at(0).addPageContents(
+	std::vector<QPDFPageObjectHelper> pages =
+            QPDFPageDocumentHelper(pdf).getAllPages();
+        QPDFPageObjectHelper& ph(pages.at(0));
+	ph.addPageContents(
             QPDFObjectHandle::newStream(
                 &pdf, "BT /F1 12 Tf 72 620 Td (Baked) Tj ET\n"), true);
-	pages.at(0).addPageContents(
+	ph.addPageContents(
             QPDFObjectHandle::newStream(
                 &pdf, "BT /F1 18 Tf 72 520 Td (Mashed) Tj ET\n"), false);
 
@@ -887,6 +900,7 @@ void runtest(int n, char const* filename1, char const* arg2)
         assert(pages.size() == 10);
         QPDFObjectHandle page5 = pages.at(5);
         pdf.removePage(page5);
+        assert(pages.size() == 9);
         pdf.addPage(page5, false);
         assert(pages.size() == 10);
         assert(pages.back().getObjGen() == page5.getObjGen());
@@ -899,10 +913,11 @@ void runtest(int n, char const* filename1, char const* arg2)
     else if (n == 19)
     {
         // Remove a page and re-insert it in the same file.
-        std::vector<QPDFObjectHandle> const& pages = pdf.getAllPages();
+        QPDFPageDocumentHelper dh(pdf);
+        std::vector<QPDFPageObjectHelper> pages = dh.getAllPages();
 
         // Try to insert a page that's already there.
-        pdf.addPage(pages.at(5), false);
+        dh.addPage(pages.at(5), false);
         std::cout << "you can't see this" << std::endl;
     }
     else if (n == 20)
@@ -932,16 +947,18 @@ void runtest(int n, char const* filename1, char const* arg2)
     else if (n == 22)
     {
         // Try to remove a page we don't have
-        std::vector<QPDFObjectHandle> const& pages = pdf.getAllPages();
-        QPDFObjectHandle page = pages.at(0);
-        pdf.removePage(page);
-        pdf.removePage(page);
+        QPDFPageDocumentHelper dh(pdf);
+        std::vector<QPDFPageObjectHelper> pages = dh.getAllPages();
+        QPDFPageObjectHelper& page = pages.at(0);
+        dh.removePage(page);
+        dh.removePage(page);
         std::cout << "you can't see this" << std::endl;
     }
     else if (n == 23)
     {
-        std::vector<QPDFObjectHandle> const& pages = pdf.getAllPages();
-        pdf.removePage(pages.back());
+        QPDFPageDocumentHelper dh(pdf);
+        std::vector<QPDFPageObjectHelper> pages = dh.getAllPages();
+        dh.removePage(pages.back());
     }
     else if (n == 24)
     {
@@ -1053,7 +1070,7 @@ void runtest(int n, char const* filename1, char const* arg2)
         newpdf.processFile(arg2);
         QPDFObjectHandle qtest = pdf.getTrailer().getKey("/QTest");
         QPDFObjectHandle O3 = qtest.getKey("/O3");
-        newpdf.addPage(O3, false);
+        QPDFPageDocumentHelper(newpdf).addPage(O3, false);
         newpdf.getTrailer().replaceKey(
             "/QTest", newpdf.copyForeignObject(qtest));
 
@@ -1073,8 +1090,9 @@ void runtest(int n, char const* filename1, char const* arg2)
         newpdf.processFile(arg2);
         QPDFObjectHandle qtest = pdf.getTrailer().getKey("/QTest");
         QPDFObjectHandle O3 = qtest.getKey("/O3");
-        newpdf.addPage(O3.getKey("/OtherPage"), false);
-        newpdf.addPage(O3, false);
+        QPDFPageDocumentHelper dh(newpdf);
+        dh.addPage(O3.getKey("/OtherPage"), false);
+        dh.addPage(O3, false);
         newpdf.getTrailer().replaceKey(
             "/QTest", newpdf.copyForeignObject(qtest));
 
@@ -1320,11 +1338,12 @@ void runtest(int n, char const* filename1, char const* arg2)
     else if (n == 37)
     {
         // Parse content streams of all pages
-        std::vector<QPDFObjectHandle> pages = pdf.getAllPages();
-        for (std::vector<QPDFObjectHandle>::iterator iter = pages.begin();
+        std::vector<QPDFPageObjectHelper> pages =
+            QPDFPageDocumentHelper(pdf).getAllPages();
+        for (std::vector<QPDFPageObjectHelper>::iterator iter = pages.begin();
              iter != pages.end(); ++iter)
         {
-            QPDFObjectHandle page = *iter;
+            QPDFPageObjectHelper& page(*iter);
             ParserCallbacks cb;
             page.parsePageContents(&cb);
         }
@@ -1341,9 +1360,10 @@ void runtest(int n, char const* filename1, char const* arg2)
     else if (n == 39)
     {
         // Display image filter and color set for each image on each page
-        std::vector<QPDFObjectHandle> pages = pdf.getAllPages();
+        std::vector<QPDFPageObjectHelper> pages =
+            QPDFPageDocumentHelper(pdf).getAllPages();
         int pageno = 0;
-        for (std::vector<QPDFObjectHandle>::iterator p_iter =
+        for (std::vector<QPDFPageObjectHelper>::iterator p_iter =
                  pages.begin();
              p_iter != pages.end(); ++p_iter)
         {
@@ -1378,8 +1398,9 @@ void runtest(int n, char const* filename1, char const* arg2)
     {
         // Apply a token filter. This test case is crafted to work
         // with coalesce.pdf.
-        std::vector<QPDFObjectHandle> pages = pdf.getAllPages();
-        for (std::vector<QPDFObjectHandle>::iterator iter =
+        std::vector<QPDFPageObjectHelper> pages =
+            QPDFPageDocumentHelper(pdf).getAllPages();
+        for (std::vector<QPDFPageObjectHelper>::iterator iter =
                  pages.begin();
              iter != pages.end(); ++iter)
         {
@@ -1446,9 +1467,150 @@ void runtest(int n, char const* filename1, char const* arg2)
         std::cerr << "One error\n";
         array.getArrayItem(1).getKey("/K").getArrayItem(0).getStringValue();
         // Stream dictionary
-        QPDFObjectHandle page = pdf.getAllPages()[0];
+        QPDFObjectHandle page = pdf.getAllPages().at(0);
         assert("/QPDFFakeName" ==
                page.getKey("/Contents").getDict().getKey("/Potato").getName());
+        // Rectangles
+        QPDFObjectHandle::Rectangle r0 = integer.getArrayAsRectangle();
+        assert((r0.llx == 0) && (r0.lly == 0) &&
+               (r0.urx == 0) && (r0.ury == 0));
+        QPDFObjectHandle rect = QPDFObjectHandle::newFromRectangle(
+            QPDFObjectHandle::Rectangle(1.2, 3.4, 5.6, 7.8));
+        QPDFObjectHandle::Rectangle r1 = rect.getArrayAsRectangle();
+        assert((r1.llx > 1.19) && (r1.llx < 1.21) &&
+               (r1.lly > 3.39) && (r1.lly < 3.41) &&
+               (r1.urx > 5.59) && (r1.urx < 5.61) &&
+               (r1.ury > 7.79) && (r1.ury < 7.81));
+    }
+    else if (n == 43)
+    {
+        // Forms
+        QPDFAcroFormDocumentHelper afdh(pdf);
+        if (! afdh.hasAcroForm())
+        {
+            std::cout << "no forms\n";
+            return;
+        }
+        std::cout << "iterating over form fields\n";
+        std::vector<QPDFFormFieldObjectHelper> form_fields =
+            afdh.getFormFields();
+        for (std::vector<QPDFFormFieldObjectHelper>::iterator iter =
+                 form_fields.begin();
+             iter != form_fields.end(); ++iter)
+        {
+            QPDFFormFieldObjectHelper ffh(*iter);
+            std::cout << "Field: " << ffh.getObjectHandle().unparse()
+                      << std::endl;
+            QPDFFormFieldObjectHelper node = ffh;
+            while (! node.isNull())
+            {
+                QPDFFormFieldObjectHelper parent(node.getParent());
+                std::cout << "  Parent: "
+                          << (parent.isNull()
+                              ? "none"
+                              : parent.getObjectHandle().unparse())
+                          << std::endl;
+                node = parent;
+            }
+            std::cout << "  Fully qualified name: "
+                      << ffh.getFullyQualifiedName() << std::endl;
+            std::cout << "  Partial name: "
+                      << ffh.getPartialName() << std::endl;
+            std::cout << "  Alternative name: "
+                      << ffh.getAlternativeName() << std::endl;
+            std::cout << "  Mapping name: "
+                      << ffh.getMappingName() << std::endl;
+            std::cout << "  Field type: "
+                      << ffh.getFieldType() << std::endl;
+            std::cout << "  Value: "
+                      << ffh.getValue().unparse() << std::endl;
+            std::cout << "  Value as string: "
+                      << ffh.getValueAsString() << std::endl;
+            std::cout << "  Default value: "
+                      << ffh.getDefaultValue().unparse() << std::endl;
+            std::cout << "  Default value as string: "
+                      << ffh.getDefaultValueAsString() << std::endl;
+            std::cout << "  Default appearance: "
+                      << ffh.getDefaultAppearance() << std::endl;
+            std::cout << "  Quadding: "
+                      << ffh.getQuadding() << std::endl;
+            std::vector<QPDFAnnotationObjectHelper> annotations =
+                afdh.getAnnotationsForField(ffh);
+            for (std::vector<QPDFAnnotationObjectHelper>::iterator i2 =
+                     annotations.begin();
+                 i2 != annotations.end(); ++i2)
+            {
+                std::cout << "  Annotation: "
+                          << (*i2).getObjectHandle().unparse() << std::endl;
+            }
+        }
+        std::cout << "iterating over annotations per page\n";
+        std::vector<QPDFPageObjectHelper> pages =
+            QPDFPageDocumentHelper(pdf).getAllPages();
+        for (std::vector<QPDFPageObjectHelper>::iterator iter = pages.begin();
+             iter != pages.end(); ++iter)
+        {
+            std::cout << "Page: " << (*iter).getObjectHandle().unparse()
+                      << std::endl;
+            std::vector<QPDFAnnotationObjectHelper> annotations =
+                afdh.getWidgetAnnotationsForPage(*iter);
+            for (std::vector<QPDFAnnotationObjectHelper>::iterator i2 =
+                     annotations.begin();
+                 i2 != annotations.end(); ++i2)
+            {
+                QPDFAnnotationObjectHelper ah(*i2);
+                std::cout << "  Annotation: " << ah.getObjectHandle().unparse()
+                          << std::endl;
+                std::cout << "    Field: "
+                          << (afdh.getFieldForAnnotation(ah).
+                              getObjectHandle().unparse())
+                          << std::endl;
+                std::cout << "    Subtype: " << ah.getSubtype() << std::endl;
+                std::cout << "    Rect: ";
+                print_rect(std::cout, ah.getRect());
+                std::cout << std::endl;
+                std::string state = ah.getAppearanceState();
+                if (! state.empty())
+                {
+                    std::cout << "    Appearance state: " << state
+                              << std::endl;
+                }
+                std::cout << "    Appearance stream (/N): "
+                          << ah.getAppearanceStream("/N").unparse()
+                          << std::endl;
+                std::cout << "    Appearance stream (/N, /3): "
+                          << ah.getAppearanceStream("/N", "/3").unparse()
+                          << std::endl;
+            }
+        }
+    }
+    else if (n == 44)
+    {
+        // Set form fields.
+        QPDFAcroFormDocumentHelper afdh(pdf);
+        std::vector<QPDFFormFieldObjectHelper> fields = afdh.getFormFields();
+        for (std::vector<QPDFFormFieldObjectHelper>::iterator iter =
+                 fields.begin();
+             iter != fields.end(); ++iter)
+        {
+            QPDFFormFieldObjectHelper& field(*iter);
+            QPDFObjectHandle ft = field.getInheritableFieldValue("/FT");
+            if (ft.isName() && (ft.getName() == "/Tx"))
+            {
+                // \xc3\xb7 is utf-8 for U+00F7 (divided by)
+                field.setV("3.14 \xc3\xb7 0");
+                std::cout << "Set field value: "
+                          << field.getFullyQualifiedName()
+                          << " -> "
+                          << field.getValueAsString()
+                          << std::endl;
+            }
+        }
+        QPDFWriter w(pdf, "a.pdf");
+	w.setQDFMode(true);
+        w.setStaticID(true);
+        w.setSuppressOriginalObjectIDs(true);
+        w.write();
     }
     else
     {
